import argparse
import collections
import copy
import functools
import ipaddress
import os
from pprint import pprint as pp
import re
import six
import sys
import yaml

if sys.version_info[0] == 3:
    unicode = lambda x: x


##
# A hack needed to save order of fields in output. The only reason it is here
# for verbatim reproduction of a part of original reclass.storage formula.
# It could be omitted, but the order of keys would become different.
class UnsortableList(list):
    def sort(self, *a, **k):
        pass

class UnsortableDict(collections.OrderedDict):
    def items(self, *a, **k):
        return UnsortableList(collections.OrderedDict.items(self, *a, **k))

yaml.add_representer(UnsortableDict,
                     yaml.representer.SafeRepresenter.represent_dict)
# End hack
##


_param_reference = re.compile('\$\{_param\:.*?\}')
_refernce_name = re.compile('\$\{_param\:(.*?)\}')
_iprange = re.compile('<<(.*?)>>')
NODEF = 'NOTADEFINITION'
CTRAIT = ['storage.system', 'cluster.']
STRAIT = 'system/reclass/storage'
ITRAIT = 'infra/init.yml'


def get_references(string):
    """Exctracts all raw references from a string."""
    return re.findall(_param_reference, string)


def get_ref_names(string):
    """Exctracts all reference values from a string."""
    return re.findall(_refernce_name, string)


def contains(x, *sub):
    return any(x.find(s) != -1 for s in sub)


def get_params(data):
    return data.get('parameters', {}).get('_param', {})


def get_nodes(data):
    return data.get('parameters', {}).get('reclass', {}).get(
                    'storage', {}).get('node', {})


def get_cls(data):
    return data.get('classes', [])


def get_system(data):
    return data.get("parameters", {}).get("linux", {}).get("system", {})


def has_subst(value):
    """Checks if an element has slot for substitution."""
    # NOTE: the code does not address possible nested references so far since
    # it is unclear if anyone actually uses them.
    if type(value) is str:
        return len(get_references(value)) > 0
    elif isinstance(value, collections.Iterable):
        return any(has_subst(x) for x in value)
    elif isinstance(value, dict):
        return any(has_subst(k) or has_subst(v) for k, v in six.iteritems(value))
    return False


def do_substitution(entity, src, subst_stub=None):
    """Attempts to substitute references in entity from src."""
    subster = functools.partial(do_substitution, src=src)
    if type(entity) is str:
        lookfor = zip(get_ref_names(entity), get_references(entity))
        result = entity
        for subst, ref in lookfor:
            default_subst = subst if subst_stub is None else subst_stub
            result = result.replace(ref, str(src.get(subst, default_subst)))
        return result
    elif type(entity) is list:
        return [subster(x) for x in entity]
    elif isinstance(entity, dict):
        return dict((subster(k), subster(v)) for k, v in six.iteritems(entity))
    return entity


# Parameters as they arrive after collection can easily contain references, but
# fortunately most of them could be resolved using the very same dictionary:
# since it is a glob of paramteres from different parts of a cluster model it
# can contain them. Thus it is worthwhile to try and expand parameters.
def expand_params(source):
    """Expands parameter dict in case parameters contain references."""
    result = do_substitution(source, source)
    if has_subst(result):
        result = expand_params(result)
    return result


# TODO: remove side effects.
def update_dict(x, y, block_override=False):
    for key, value in six.iteritems(y):
        if key in x and x[key] != value:
            if (type(x[key]) == type(value) and type(value) == str
                    and block_override):
                continue
            # NOTE: the substitution rules should be moved.
            # Don't replace specific value with a reference.
            if not has_subst(x[key]) and has_subst(value):
                continue
            # Do not replace a dict with a scalar value.
            if isinstance(x[key], dict) and not isinstance(value, dict):
                continue
            # Do not touch lists: it happens in other place.
            if type(x[key]) == list:
            # TODO: this is a crude cludge to patch missing class problem.  It
            # will be used till more statistics is collected and a generic
            # solution is created.
                if key == 'classes':
                    [x[key].extend(value) for el in value if 'gateway' in el]
                continue
            if isinstance(x[key], dict) and isinstance(value, dict):
                update_dict(x[key], value)
                continue
        x[key] = value


def get_substitution_parameters(source):
    exps, result = [get_params(x) for x in source], {}
    [update_dict(result, e) for e in exps]
    return result


external = {}
def get_configs(base, cluster):
    configs_to_process, out = [], []
    cluster_def = os.path.join(base, cluster)
    for rootdir, _, files in os.walk(cluster_def):
        for fname in [x for x in files if x.endswith('yml')]:
            config = os.path.join(rootdir, fname)
            if fname == 'init.yml':
                if contains(config, ITRAIT):
                    with open(config, 'r') as f:
                        external.update(get_system(yaml.safe_load(f)))
                configs_to_process.append(config)
            # NOTE: this is a special case left here for the time being.
            elif fname == 'nodes.yml':  # TODO: refactor it.
                nodes_definition = config
            else:
                with open(config, 'r') as f:
                    data = yaml.safe_load(f)
                    if data is None:
                        continue
                    if get_params(data):
                        configs_to_process.append(config)
    for config in configs_to_process:
        with open(config, 'r') as f:
            data = yaml.safe_load(f)
            data['src'] = [config]
            out.append(data)
    return out, nodes_definition


def get_basic_node_defenitions(cfgs, params, basename):
    def fixname(x):
        fn = os.path.join(basename, x.replace('.', '/') + '.yml')
        return fn if os.path.isfile(fn) else os.path.join(fn[:-4], 'init.yml')

    def innerread(x, params, content):
        with open(x, 'r') as f:
            data = yaml.safe_load(f)
        params = {} if params is None else params
        data = {} if data is None else data
        update_dict(params, get_params(data), True)
        update_dict(content, get_nodes(data),  True)
        depends_on = get_cls(data)
        # TODO: this works around single (so far) issue with structured
        # reference in system/reclass/storage.
        if depends_on and contains(x, STRAIT):
            for klass in depends_on:
                innerread(fixname(klass), params, content)
        return data

    out, storage_cnames, cluster_cn = collections.defaultdict(dict), [], []
    for cfg in cfgs:
        storage_cnames.extend(x for x in get_cls(cfg) if contains(x, *CTRAIT))
    for x in map(fixname, storage_cnames):
        node_content = {}
        data = innerread(x, params, node_content)
        for nodename, nodecontent in six.iteritems(node_content):
            if out[nodename].get('src') is not None:
                out[nodename]['src'].append(x)
            else:
                out[nodename]['src'] = [x]
            update_dict(out[nodename], nodecontent, False)
            # NOTE: this fixes cfg01 till its definition starts confirming
            #       to rules observed for other nodes.
            if 'cfg01' in nodename:
                out[nodename].update({'name': external['name'],
                                      'domain': external['domain']})
    return out


def dump_to_files(content, dest):
    for res in content:
        tt = ".".join([res.get('name', NODEF), res.get('domain',NODEF), 'yml'])
        if contains(tt, NODEF):
            continue
        systemdesc = UnsortableDict()
        systemdesc['name'] = res.get('name')
        systemdesc['domain'] = res.get('domain')
        # NOTE: this should stay here until cfg01 definition stabilizes.
        if systemdesc['name'] != 'cfg01':
            systemdesc['cluster'] = res.get('cluster', 'default')
            systemdesc['environment'] = res.get('environment', 'prd')
        params = res.get('params', {})
        if params:
            contents = {'classes': res['classes'],
                        'parameters': {
                          '_param': params,
                          'linux': {
                             'system':  systemdesc
            }}}
        # NOTE: this should stay here until cfg01 definition stabilizes.
        else:
            contents = {'classes': res['classes'],
                        'parameters': {
                          'linux': {
                             'system':  systemdesc
            }}}
        # NOTE: the original formula contains hints to other possible sections.
        # Since it is not immediately clear whether those are used or just
        # a preacution for future work thay are omited for now.
        with open(os.path.join(dest, tt), 'w') as f:
            yaml.dump(contents, f, default_flow_style=False)



# Takes an internally parametrized node desciption and generates appropriate
# number of nodes.
# A node consists of a regular node definition fields and additional 'repeat'
# field which is used to update/populate regular fields. Parammetrized fields
# are apparently of two types: number and ip address, substitution is defined
# by adding double angle brackets around a name. Name 'count' is apparently
# hardcoded for various counter-based name parts, names for ip ranges are
# somewhat dinamic.
def do_count_expansion(node_description):
    def get_ip(ranges):
        af = lambda x: ipaddress.ip_address(unicode(x))
        for sta, sto in [x.split('-') for x in ranges.split(',')]:
            ista, isto = af(sta), af(sto)
            while ista <= isto:
                yield ista
                ista += 1

    def generate_counted_node(from_node, count_padding):
        def update_count(string, value, padding):
            return string.replace('<<count>>', str(value).zfill(padding))
        out = {}
        for k, v in six.iteritems(from_node):
            if type(v) is str:
                outv = update_count(v, count, count_padding)
            elif type(v) is list:
                outv = v
            # Tricky place: a dictionary can contain ip range substitution
            # TODO: refactor.
            elif type(v) is dict:
                outv = copy.deepcopy(v)
                for ik, iv in six.iteritems(v):
                    if type(iv) is dict:
                        # apparently key is always 'value'
                        for iiv in six.itervalues(iv):
                            ranges = re.findall(_iprange, iiv)
                            if ranges:
                                addr = six.next(ip_ranges[ranges[0]])
                                outv[ik] = str(addr.exploded)
            out[k] = outv
        return out

    local_node, result = copy.deepcopy(node_description), []
    # Process repeat field
    repeat = local_node.pop('repeat')
    local_node['params'].update(repeat['params'])
    # Process ranges. Ranges look like the only repeatable entity, so
    # it is not generalized yet. NOTE: check other models.
    ip_ranges = six.iteritems(repeat['ip_ranges'])
    ip_ranges = dict((ip_rn, get_ip(ip_r)) for ip_rn, ip_r in ip_ranges)
    # Generate counted nodes.
    for count in range(repeat['start'], repeat['start'] + repeat['count']):
        result.append(generate_counted_node(local_node, repeat['digits']))
    return result


def extend_basic_nodes(basic_nodes, nodes_definition):
    # Here nodes defined in nodes.yml are transformed into actual definitions.
    def needs_expansion(node):
        return 'repeat' in node.keys()

    with open(nodes_definition, 'r') as f:
        nodes_init = yaml.safe_load(f)['parameters']['reclass']['storage']['node']
    basic_nodes = copy.deepcopy(basic_nodes)  # preserve state.
    basic_nodes_extended, extra_nodes = {}, []

    for nname, node in six.iteritems(basic_nodes):
        if needs_expansion(node):
            continue
        extension = nodes_init.get(nname, {}).get('classes', [])
        to_extend = node.get('classes', [])#.extend(extension)
        if to_extend != extension:
            to_extend.extend(extension)
        node.get('params', {}).update(nodes_init.get(nname, {}).
                 get('params', {}))
        basic_nodes_extended[nname] = node

    for node in nodes_init.values():
        if needs_expansion(node):
            extra_nodes.extend(do_count_expansion(node))
    all_nodes = list(basic_nodes_extended.values())
    all_nodes.extend(extra_nodes)
    return all_nodes


def main():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-b', '--basedir', default='/srv/salt/reclass/classes',
                        help="Cluster base directory")
    parser.add_argument('cluster', help="Cluster name")
    parser.add_argument('-o', '--outdir', default='_newgen',
                        help="Directory to store generated nodes")
    args = parser.parse_args()

    cluster_name = os.path.join('cluster', args.cluster)
    configs, nodes_definition = get_configs(args.basedir, cluster_name)
    params = get_substitution_parameters(configs)
    basic_nodes = get_basic_node_defenitions(configs, params, args.basedir)
    all_nodes = extend_basic_nodes(basic_nodes, nodes_definition)
    expanded_parameters = expand_params(params)
    resulting_nodes = do_substitution(all_nodes, expanded_parameters)
    dump_to_files(resulting_nodes, args.outdir)


if __name__ == '__main__':
    main()
